use crate::read;
use itertools::Itertools;

#[derive(Debug, PartialEq, Eq, Clone, Hash)]
pub enum Content {
    Number(i32),
    List(Vec<Content>),
}

#[derive(Debug, PartialEq, Eq, Clone, Hash, PartialOrd, Ord)]
pub struct Packet {
    pub content: Vec<Content>,
}

#[derive(Debug, PartialEq, Eq, Clone, Hash)]
pub struct Pair {
    pub first: Packet,
    pub second: Packet,
}

impl Pair {
    pub fn in_order(&self) -> bool {
        self.first < self.second
    }
}

impl Ord for Content {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        match (self, other) {
            (Content::Number(a), Content::Number(b)) => a.cmp(b),
            (Content::Number(a), Content::List(b)) => vec![Content::Number(*a)].cmp(b),
            (Content::List(a), Content::Number(b)) => a.cmp(&vec![Content::Number(*b)]),
            (Content::List(a), Content::List(b)) => a.cmp(b),
        }
    }
}
impl PartialOrd for Content {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

pub fn run() {
    let input = read("13");

    #[cfg(feature = "part1")]
    {
        let p1 = input
            .split("\n\n")
            .map(|pair| pair.split_once('\n').unwrap())
            .map(|pair_lines| Pair {
                first: packet_parser::packet(pair_lines.0).unwrap(),
                second: packet_parser::packet(pair_lines.1).unwrap(),
            })
            .enumerate()
            .filter(|(_, pair)| pair.in_order())
            .map(|(idx, _)| idx + 1)
            .sum::<usize>();
        println!("Day 12, Part 01: {:?}", p1);
    }

    #[cfg(feature = "part2")]
    {
        let div1 = packet_parser::packet("[[2]]").unwrap();
        let div2 = packet_parser::packet("[[6]]").unwrap();
        let mut p2 = input
            .lines()
            .filter(|l| !l.is_empty())
            .map(|l| packet_parser::packet(l).unwrap())
            .collect_vec();
        p2.push(div1.clone());
        p2.push(div2.clone());
        p2.sort_unstable();
        let product =
            (p2.binary_search(&div1).unwrap() + 1) * (p2.binary_search(&div2).unwrap() + 1);

        println!("Day 12, Part 02: {:?}", product);
    }
}

peg::parser! {
  grammar packet_parser() for str {
    pub rule packet() -> Packet
        = content:content() { Packet { content }}

    rule value() -> Content
        = content:content() { Content::List(content) } / int:int() { Content::Number(int) }

    rule int() -> i32
        = n:$(['0'..='9']+) {? n.parse().or(Err("u8")) }

    rule content() -> Vec<Content>
        = "[" content:(value() ** ",") "]" { content }
  }
}
